"""Module wrapping interactions with the local executor packages."""


import shutil
from pathlib import Path
from typing import Tuple, Optional

from .helper import unpack_package

from . import JINA_HUB_ROOT


def get_dist_path(uuid: str, tag: str) -> Tuple['Path', 'Path']:
    """Get the package path according ID and TAG
    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: package and its dist-info path
    """
    pkg_path = JINA_HUB_ROOT / uuid
    pkg_dist_path = JINA_HUB_ROOT / f'{uuid}-{tag}.dist-info'
    return pkg_path, pkg_dist_path


def install_local(
    zip_package: 'Path', uuid: str, tag: str, force: Optional[bool] = False
):
    """Install the package in zip format to the Jina Hub root.

    :param zip_package: the path of the zip file
    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :param force: if set, overwrites the package
    """

    pkg_path, pkg_dist_path = get_dist_path(uuid, tag)
    if pkg_dist_path.exists() and not force:
        return

    # clean existed dist-info
    for dist in JINA_HUB_ROOT.glob(f'{uuid}-*.dist-info'):
        shutil.rmtree(dist)
    if pkg_path.exists():
        shutil.rmtree(pkg_path)

    # unpack the zip package to the root pkg_path
    unpack_package(zip_package, pkg_path)

    # TODO: install the dependencies included in requirements.txt

    # create dist-info folder
    pkg_dist_path.mkdir(parents=False, exist_ok=True)

    manifest_path = pkg_path / 'manifest.yml'
    if manifest_path.exists():
        shutil.copyfile(manifest_path, pkg_dist_path / 'manifest.yml')

    requirements_path = pkg_path / 'requirements.txt'
    if requirements_path.exists():
        shutil.copyfile(requirements_path, pkg_dist_path / 'requirements.txt')


def uninstall_local(uuid: str):
    """Uninstall the executor package.

    :param uuid: the UUID of the executor
    """
    pkg_path, _ = get_dist_path(uuid, None)
    for dist in JINA_HUB_ROOT.glob(f'{uuid}-*.dist-info'):
        shutil.rmtree(dist)
    if pkg_path.exists():
        shutil.rmtree(pkg_path)


def list_local():
    """List the locally-available executor packages.

    :return: the list of local executors (if found)
    """
    result = []
    for dist_name in JINA_HUB_ROOT.glob(r'*-v*.dist-info'):
        result.append(dist_name)

    return result


def resolve_local(uuid: str, tag: Optional[str] = None) -> 'Path':
    """Return the path of the executor if available.

    :param uuid: the UUID of executor
    :param tag: the TAG of executor
    :return: the path of the executor package
    """
    pkg_path = JINA_HUB_ROOT / uuid
    pkg_dist_path = JINA_HUB_ROOT / f'{uuid}-{tag}.dist-info'
    if not pkg_path.exists():
        return None
    if tag and (not pkg_dist_path.exists()):
        return None
    return pkg_path


def exist_local(uuid: str, tag: str = None) -> bool:
    """Check whether the executor exists in local

    :param uuid: the UUID of the executor
    :param tag: the TAG of the executor
    :return: True if existed, else False
    """

    return resolve_local(uuid, tag=tag) is not None



def list_all_runtimes():
    """List all public runtimes that can be used directly with :class:`jina.peapods.peas.BasePea`

    # noqa: DAR101
    # noqa: DAR201
    """
    from ...peapods.runtimes.base import BaseRuntime
    from .asyncio.grpc import GRPCRuntime
    from .asyncio.http import HTTPRuntime
    from .asyncio.websocket import WebSocketRuntime
    from .container import ContainerRuntime
    from .jinad import JinadRuntime
    from .zmq.zed import ZEDRuntime

    return [
        k
        for k, s in locals().items()
        if isinstance(s, type) and issubclass(s, BaseRuntime)
    ]


def get_runtime(name: str):
    """Get a public runtime by its name

    # noqa: DAR101
    # noqa: DAR201
    """
    from ...peapods.runtimes.base import BaseRuntime
    from .asyncio.grpc import GRPCRuntime
    from .asyncio.http import HTTPRuntime
    from .asyncio.websocket import WebSocketRuntime
    from .container import ContainerRuntime
    from .jinad import JinadRuntime
    from .zmq.zed import ZEDRuntime

    s = locals()[name]
    if isinstance(s, type) and issubclass(s, BaseRuntime):
        return s
    else:
        raise TypeError(f'{s!r} is not in type {BaseRuntime!r}')



from enum import Enum
from typing import List

from jina.enums import RemoteWorkspaceState as WorkspaceState
from ..helper import classproperty


class DaemonEnum(str, Enum):
    """Base class of Enum in JinaD"""

    @classproperty
    def values(cls) -> List:
        """Get all values in Enum

        :return: Get all values in Enum
        """
        return list(map(lambda c: c.value, cls))


class IDLiterals(DaemonEnum):
    """Enum representing all values allowed in DaemonID"""

    JPOD = 'jpod'
    JPEA = 'jpea'
    JFLOW = 'jflow'
    JNETWORK = 'jnetwork'
    JWORKSPACE = 'jworkspace'


class DaemonBuild(DaemonEnum):
    """Enum representing build value passed in .jinad file"""

    DEVEL = 'devel'
    DEFAULT = 'default'
    CPU = 'default'
    GPU = 'gpu'

    # TODO (Deepankar): remove this once default becomes default
    @classproperty
    def default(cls) -> str:
        """Get default value for DaemonBuild

        :return: default value for DaemonBuild"""
        return cls.DEVEL


class PythonVersion(DaemonEnum):
    """Enum representing python versions allowed in .jinad file"""

    PY37 = '3.7'
    PY38 = '3.8'
    PY39 = '3.9'

    @classproperty
    def default(cls) -> str:
        """Get default value for PythonVersion

        :return: default value for PythonVersion"""
        return cls.PY38


class PartialDaemonModes(DaemonEnum):
    """Enum representing partial daemon modes"""

    PEA = 'pea'
    POD = 'pod'
    FLOW = 'flow'


class UpdateOperation(DaemonEnum):
    """
    Represents the type of operation to perform in the update
    We consider these an `update` operation since they **change** the underlying state
    """

    ROLLING_UPDATE = 'rolling_update'



import argparse
import asyncio
import copy
import json
from argparse import Namespace
from contextlib import ExitStack
from typing import Optional, Sequence, Dict

import os

from .... import __default_host__, __resources_path__
from ....enums import replace_enum_to_str
from ....importer import ImportExtensions
from ....jaml.helper import complete_path
from ....logging.logger import JinaLogger


class DaemonClient:
    """
    Jina Daemon client.

    :param host: the host address of ``jinad`` instance
    :param port: the port number of ``jinad`` instance
    :param logger: Jinalogger to log information.
    :param timeout: stop waiting for a response after a given number of seconds with the timeout parameter.
    """

    kind = 'pea'  # select from pea/pod, TODO: enum

    def __init__(
        self,
        host: str,
        port: int,
        logger: 'JinaLogger' = None,
        timeout: int = None,
        **kwargs,
    ):
        self.logger = logger or JinaLogger(host)
        self.timeout = timeout
        base_url = f'{host}:{port}'
        rest_url = f'http://{base_url}'
        self.alive_api = f'{rest_url}/'
        if self.kind not in ['pea', 'pod', 'workspace']:
            raise ValueError(f'daemon kind {self.kind} is not supported')
        self.store_api = f'{rest_url}/{self.kind}s'
        self.logstream_api = f'ws://{base_url}/logstream'

    @property
    def alive(self) -> bool:
        """
        Return True if ``jinad`` is alive at remote

        :return: True if ``jinad`` is alive at remote else false
        """
        with ImportExtensions(required=True):
            import requests

        try:
            r = requests.get(url=self.alive_api, timeout=self.timeout)
            return r.status_code == requests.codes.ok
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'remote manager is not alive: {ex!r}')
            return False

    def get(self, identity: str) -> Dict:
        """
        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    def post(self, *args, **kwargs) -> Dict:
        """
        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    def delete(self, *args, **kwargs) -> str:
        """
        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    def _daemonize_id(self, id: str, kind: str = 'workspace') -> str:
        return f'j{kind}-{id}'

    async def logstream(self, id: str):
        """Websocket log stream from remote Pea / Pod

        :param id:  the identity of that Pea / Pod
        """
        with ImportExtensions(required=True):
            import websockets

        remote_log_config = os.path.join(__resources_path__, 'logging.remote.yml')
        all_remote_loggers = {}
        try:
            url = f'{self.logstream_api}/{id}'
            async with websockets.connect(url) as websocket:
                async for log_line in websocket:
                    try:
                        ll = json.loads(log_line)
                        name = ll['name']
                        if name not in all_remote_loggers:
                            all_remote_loggers[name] = JinaLogger(
                                context=ll['host'], log_config=remote_log_config
                            )

                        all_remote_loggers[name].info(
                            '{host} {name} {type} {message}'.format_map(ll)
                        )
                    except json.decoder.JSONDecodeError:
                        continue
        except websockets.exceptions.ConnectionClosedOK:
            self.logger.warning(f'log streaming is disconnected')
        except websockets.exceptions.WebSocketException as e:
            self.logger.error(
                f'log streaming is disabled, you won\'t see logs on the remote\n Reason: {e!r}'
            )
        except asyncio.CancelledError:
            self.logger.info(f'log streaming is cancelled')
        finally:
            for l in all_remote_loggers.values():
                l.close()

    def _mask_args(self, args: 'argparse.Namespace'):
        _args = copy.deepcopy(args)

        # reset the runtime to ZEDRuntime or ContainerRuntime
        if _args.runtime_cls == 'JinadRuntime':
            if _args.uses.startswith('docker://'):
                _args.runtime_cls = 'ContainerRuntime'
            else:
                _args.runtime_cls = 'ZEDRuntime'

        # TODO:/NOTE this prevents jumping from remote to another remote (Han: 2021.1.17)
        # _args.host = __default_host__
        # host resetting disables dynamic routing. Use `disable_remote` instead
        _args.disable_remote = True

        # NOTE: on remote relative filepaths should be converted to filename only
        def basename(field):
            if field and not field.startswith('docker://'):
                try:
                    return os.path.basename(complete_path(field))
                except FileNotFoundError:
                    pass
            return field

        for f in ('uses', 'uses_after', 'uses_before', 'py_modules'):
            attr = getattr(_args, f, None)
            if not attr:
                continue
            setattr(_args, f, [basename(m) for m in attr]) if isinstance(
                attr, list
            ) else setattr(_args, f, basename(attr))

        _args.log_config = ''  # do not use local log_config
        _args.upload_files = []  # reset upload files
        _args.noblock_on_start = False  # wait until start success

        changes = []
        for k, v in vars(_args).items():
            if v != getattr(args, k):
                changes.append(f'{k:>30s}: {str(getattr(args, k)):30s} -> {str(v):30s}')
        if changes:
            changes = [
                'note the following arguments have been masked or altered for remote purpose:'
            ] + changes
            self.logger.warning('\n'.join(changes))

        return _args


class PeaDaemonClient(DaemonClient):
    """Pea API, we might have different endpoints for peas & pods later"""

    kind = 'pea'

    def get(self, id: str) -> Dict:
        """Get status of the remote Pea / Pod

        :param id: 'DaemonID' based identity for the Pea
        :raises: requests.exceptions.RequestException
        :return: json response of the remote Pea / Pod status
        :rtype: Dict
        """

        with ImportExtensions(required=True):
            import requests

        try:
            r = requests.get(
                url=f'{self.store_api}/{self._daemonize_id(id, self.kind)}',
                timeout=self.timeout,
            )
            if r.status_code == requests.codes.not_found:
                self.logger.info(f'couldn\'t find {id} in remote {self.kind} store')
            return r.json()
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'can\'t get status of {self.kind}: {ex!r}')

    def post(self, args: 'Namespace') -> Optional[str]:
        """Create a remote Pea / Pod

        :param args: the arguments for remote Pea
        :type args: Namespace
        :raises: requests.exceptions.RequestException
        :return: the identity of the spawned Pea / Pod
        :rtype: Optional[str]
        """

        with ImportExtensions(required=True):
            import requests

        try:
            payload = replace_enum_to_str(vars(self._mask_args(args)))
            # set timeout to None if args.timeout_ready is -1 (wait forever)
            r = requests.post(
                url=self.store_api,
                params={'workspace_id': self._daemonize_id(id=args.workspace_id)},
                json=payload,
                timeout=args.timeout_ready if args.timeout_ready != -1 else None,
            )
            rj = r.json()
            if r.status_code == 201:
                return rj
            elif r.status_code == 400:
                # known internal error
                rj_body = '\n'.join(j for j in rj['body'])
                self.logger.error(f'{rj["detail"]}\n{rj_body}')
            elif r.status_code == 422:
                self.logger.error(
                    'your payload is not correct, please follow the error message and double check'
                )
            raise requests.exceptions.RequestException(rj)
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'failed to create as {ex!r}')

    def delete(self, id: str, **kwargs) -> bool:
        """
        Delete a remote pea/pod

        :param id: the identity of that pea/pod
        :param kwargs: keyword arguments
        :return: True if the deletion is successful
        """
        with ImportExtensions(required=True):
            import requests

        try:
            r = requests.delete(url=f'{self.store_api}/{id}', timeout=self.timeout)
            rj = r.json()
            if r.status_code != requests.codes.ok:
                rj_body = '\n'.join(j for j in rj['body'])
                self.logger.error(
                    f'deletion for {id} failed: {rj["detail"]}\n{rj_body}'
                )
            return r.status_code == requests.codes.ok
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'failed to delete {id} as {ex!r}')
            return False


class PodDaemonClient(PeaDaemonClient):
    """Pod API, we might have different endpoints for peas & pods later"""

    kind = 'pod'


class WorkspaceDaemonClient(PeaDaemonClient):
    """ Defines interaction with Daemon Workspace API  """

    kind = 'workspace'

    def post(self, dependencies: Sequence[str], workspace_id: str) -> Optional[Dict]:
        """Create a remote workspace (includes file upload, docker build on remote)

        :param dependencies: file dependencies
        :param workspace_id: Workspace to which the files will get uploaded
        :raises: requests.exceptions.RequestException
        :return: dict response for upload
        """
        with ImportExtensions(required=True):
            import requests

        with ExitStack() as file_stack:
            files = [
                (
                    'files',
                    file_stack.enter_context(open(complete_path(f), 'rb')),
                )
                for f in dependencies
            ]
            try:
                if files:
                    self.logger.info(f'uploading {len(files)} file(s): {dependencies}')
                r = requests.post(
                    url=self.store_api,
                    params={'id': self._daemonize_id(id=workspace_id)},
                    files=files if files else None,
                    timeout=self.timeout,
                )
                rj = r.json()
                if r.status_code == requests.codes.created:
                    return rj
                else:
                    raise requests.exceptions.RequestException(rj)
            except requests.exceptions.RequestException as ex:
                self.logger.error(f'fail to upload as {ex!r}')

    def delete(self, id: str, **kwargs) -> bool:
        """
        Delete a remote workspace

        :param id: the identity of that pea/pod/workspace
        :param kwargs: keyword arguments
        :return: True if the deletion is successful
        """
        with ImportExtensions(required=True):
            import requests

        try:
            # NOTE: This deletes the container, network by default & leaves the files as-is on remote.
            # TODO: do we parameterize it?
            r = requests.delete(
                url=f'{self.store_api}/{self._daemonize_id(id)}',
                params={'container': True, 'network': True, 'files': False},
                timeout=self.timeout,
            )
            rj = r.json()
            if r.status_code != requests.codes.ok:
                rj_body = ''.join(j for j in rj['body'])
                self.logger.error(
                    f'deletion for {id} failed: {rj["detail"]}\n{rj_body}'
                )
            return r.status_code == requests.codes.ok
        except requests.exceptions.RequestException as ex:
            self.logger.error(f'failed to delete {id} as {ex!r}')
            return False



import argparse
import os
from typing import List


def api_to_dict():
    """Convert Jina API to a dict
    :return: dict
    """
    from jina import __version__
    from jina.parsers import get_main_parser

    parsers = get_main_parser()._actions[-1].choices

    all_d = {
        'name': 'Jina',
        'description': 'Jina is the cloud-native neural search solution powered by state-of-the-art AI and deep '
        'learning technology',
        'license': 'Apache 2.0',
        'vendor': 'Jina AI Limited',
        'source': 'https://github.com/jina-ai/jina/tree/'
        + os.environ.get('JINA_VCS_VERSION', 'master'),
        'url': 'https://jina.ai',
        'docs': 'https://docs.jina.ai',
        'authors': 'dev-team@jina.ai',
        'version': __version__,
        'methods': [],
        'revision': os.environ.get('JINA_VCS_VERSION'),
    }

    for p_name in parsers.keys():
        d = {'name': p_name, 'options': []}
        for ddd in _export_parser_args(
            lambda *x: get_main_parser()._actions[-1].choices[p_name], type_as_str=True
        ):
            d['options'].append(ddd)
        all_d['methods'].append(d)

    return all_d


def _export_parser_args(parser_fn, type_as_str: bool = False, **kwargs):
    from jina.enums import BetterEnum
    from argparse import _StoreAction, _StoreTrueAction
    from jina.parsers.helper import KVAppendAction, _SHOW_ALL_ARGS

    port_attr = ('help', 'choices', 'default', 'required', 'option_strings', 'dest')
    parser = parser_fn(**kwargs)
    parser2 = parser_fn(**kwargs)
    random_dest = set()
    for a, b in zip(parser._actions, parser2._actions):
        if a.default != b.default:
            random_dest.add(a.dest)
    for a in parser._actions:
        if isinstance(a, (_StoreAction, _StoreTrueAction, KVAppendAction)):
            if not _SHOW_ALL_ARGS and a.help == argparse.SUPPRESS:
                continue
            ddd = {p: getattr(a, p) for p in port_attr}
            if isinstance(a, _StoreTrueAction):
                ddd['type'] = bool
            elif isinstance(a, KVAppendAction):
                ddd['type'] = dict
            else:
                ddd['type'] = a.type
            if ddd['choices']:
                ddd['choices'] = [
                    str(k) if isinstance(k, BetterEnum) else k for k in ddd['choices']
                ]
                ddd['type'] = str
            if isinstance(ddd['default'], BetterEnum):
                ddd['default'] = str(ddd['default'])
                ddd['type'] = str
            if ddd['type'] == str and (a.nargs == '*' or a.nargs == '+'):
                ddd['type'] = List[str]
        else:
            continue

        if a.dest in random_dest:
            ddd['default_random'] = True
            from jina.helper import random_identity, random_port

            if isinstance(a.default, str):
                ddd['default_factory'] = random_identity.__name__
            elif isinstance(a.default, int):
                ddd['default_factory'] = random_port.__name__
        else:
            ddd['default_random'] = False

        if type_as_str:
            ddd['type'] = getattr(ddd['type'], '__name__', str(ddd['type']))
        ddd['name'] = ddd.pop('dest')
        yield ddd



import traceback

from fastapi import Request, status
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError


class DockerNotFoundException(Exception):
    """ Exception if docker client cannot communicate with dockerd"""


class DockerImageException(Exception):
    """ Exception while handling a docker image in the workspace"""


class DockerNetworkException(Exception):
    """ Exception while handling docker networks in the workspace """


class DockerContainerException(Exception):
    """ Exception while handling a docker container in the workspace"""


class Runtime400Exception(Exception):
    """Exception when daemon FastAPI app is running"""


async def daemon_runtime_exception_handler(request: Request, ex: 'Runtime400Exception'):
    """Exception handler for all Runtime400Exceptions

    :param request: starlette request
    :param ex: actual Runtime400Exception
    :return: json response representing the error
    """
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST,
        content=jsonable_encoder(
            {
                'detail': repr(ex),
                'body': traceback.format_exception(
                    etype=type(ex), value=ex, tb=ex.__traceback__
                ),
            }
        ),
    )


async def validation_exception_handler(request: Request, ex: 'RequestValidationError'):
    """Exception handler for all RequestValidationError raised by pydantic

    :param request: starlette request
    :param ex: actual Validation exception
    :return: json response representing the error
    """
    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content=jsonable_encoder({"detail": ex.errors(), "body": str(ex)}),
    )



"""Argparser module for remote runtime"""
from ...helper import add_arg_group
from .... import __default_host__
from .... import helper
from ....enums import CompressAlgo


def mixin_remote_parser(parser):
    """Add the options for remote expose
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Expose')

    gp.add_argument(
        '--host',
        type=str,
        default=__default_host__,
        help=f'The host address of the runtime, by default it is {__default_host__}.',
    )

    gp.add_argument(
        '--port-expose',
        type=int,
        default=helper.random_port(),
        help='The port of the host exposed to the public',
    )

    gp.add_argument(
        '--proxy',
        action='store_true',
        default=False,
        help='If set, respect the http_proxy and https_proxy environment variables. '
        'otherwise, it will unset these proxy variables before start. '
        'gRPC seems to prefer no proxy',
    )


def mixin_http_gateway_parser(parser=None):
    """Add the options to rest server

    :param parser: the parser
    """
    gp = add_arg_group(parser, title='HTTP Gateway')

    gp.add_argument(
        '--title',
        type=str,
        help='The title of this HTTP server. It will be used in automatics docs such as Swagger UI.',
    )

    gp.add_argument(
        '--description',
        type=str,
        help='The description of this HTTP server. It will be used in automatics docs such as Swagger UI.',
    )

    gp.add_argument(
        '--cors',
        action='store_true',
        default=False,
        help='''
        If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access. 
        ''',
    )

    gp.add_argument(
        '--no-debug-endpoints',
        action='store_true',
        default=False,
        help='If set, /status /post endpoints are removed from HTTP interface. ',
    )

    gp.add_argument(
        '--no-crud-endpoints',
        action='store_true',
        default=False,
        help='''
        If set, /index, /search, /update, /delete endpoints are removed from HTTP interface.
        
        Any executor that has `@requests(on=...)` bind with those values will receive data requests. 
        ''',
    )

    gp.add_argument(
        '--expose-endpoints',
        type=str,
        help='''
        A JSON string that represents a map from executor endpoints (`@requests(on=...)`) to HTTP endpoints.
        ''',
    )


def mixin_prefetch_gateway_parser(parser=None):
    """Add the options for prefetching
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Prefetch')

    gp.add_argument(
        '--prefetch',
        type=int,
        default=50,
        help='The number of pre-fetched requests from the client',
    )
    gp.add_argument(
        '--prefetch-on-recv',
        type=int,
        default=1,
        help='The number of additional requests to fetch on every receive',
    )


def mixin_compressor_parser(parser=None):
    """Add the options for compressors
    :param parser: the parser
    """
    gp = add_arg_group(parser, title='Compression')

    gp.add_argument(
        '--compress',
        type=CompressAlgo.from_string,
        choices=list(CompressAlgo),
        default=CompressAlgo.LZ4,
        help='''
    The compress algorithm used over the entire Flow.

    Note that this is not necessarily effective, 
    it depends on the settings of `--compress-min-bytes` and `compress-min-ratio`''',
    )

    gp.add_argument(
        '--compress-min-bytes',
        type=int,
        default=1024,
        help='The original message size must be larger than this number to trigger the compress algorithm, '
        '-1 means disable compression.',
    )

    gp.add_argument(
        '--compress-min-ratio',
        type=float,
        default=1.1,
        help='The compression ratio (uncompressed_size/compressed_size) must be higher than this number '
        'to trigger the compress algorithm.',
    )



"""Module for helper functions for Hub API."""

import io
import os
import hashlib
from typing import Tuple, Optional
import zipfile
from pathlib import Path
from urllib.parse import urlparse
from .. import __resources_path__
from ..importer import ImportExtensions


def parse_hub_uri(uri_path: str) -> Tuple[str, str, str, str]:
    """Parse the uri of the Jina Hub executor.

    :param uri_path: the uri of Jina Hub executor
    :return: a tuple of schema, id, tag, and secret
    """
    parser = urlparse(uri_path)
    scheme = parser.scheme
    items = list(parser.netloc.split(':'))
    name = items[0]
    secret = items[1] if len(items) > 1 else None
    tag = parser.path.strip('/') if parser.path else None
    return scheme, name, tag, secret


def md5file(file_path: 'Path') -> str:
    """Retrun the MD5 checksum of the file

    :param file_path: the file to check md5sum
    :return: the MD5 checksum
    """
    hash_md5 = hashlib.md5()
    with file_path.open(mode='rb') as fp:
        for chunk in iter(lambda: fp.read(128 * hash_md5.block_size), b''):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def unpack_package(filepath: 'Path', target_dir: 'Path'):
    """Unpack the file to the target_dir.

    :param filepath: the path of given file
    :param target_dir: the path of target folder
    """
    if filepath.suffix == '.zip':
        with zipfile.ZipFile(filepath, 'r') as zip:
            zip.extractall(target_dir)
    elif filepath.suffix in ['.tar', '.gz']:
        with zipfile.open(filepath) as tar:
            tar.extractall(target_dir)
    else:
        raise ValueError("File format is not supported for unpacking.")


def archive_package(package_folder: 'Path') -> 'io.BytesIO':
    """
    Archives the given folder in zip format and return a data stream.
    :param package_folder: the folder path of the package
    :return: the data stream of zip content
    """

    with ImportExtensions(required=True):
        import pathspec

    root_path = package_folder.resolve()

    gitignore = root_path / '.gitignore'
    if not gitignore.exists():
        gitignore = Path(__resources_path__) / 'Python.gitignore'

    with gitignore.open() as fp:
        ignored_spec = pathspec.PathSpec.from_lines('gitwildmatch', fp)
        ignored_spec += pathspec.PathSpec.from_lines('gitwildmatch', ['.git'])

    zip_stream = io.BytesIO()
    try:
        zfile = zipfile.ZipFile(zip_stream, 'w', compression=zipfile.ZIP_DEFLATED)
    except EnvironmentError as e:
        raise e

    def _zip(base_path, path, archive):
        for p in path.iterdir():
            if ignored_spec.match_file(p):
                continue
            if p.is_dir():
                _zip(base_path, p, archive)
            else:
                archive.write(p, p.relative_to(base_path))

    _zip(root_path, root_path, zfile)

    zfile.close()
    zip_stream.seek(0)

    return zip_stream


def download_with_resume(
    url: str,
    target_dir: 'Path',
    filename: Optional[str] = None,
    md5sum: Optional[str] = None,
) -> 'Path':
    """
    Download file from url to target_dir, and check md5sum.
    Performs a HTTP(S) download that can be restarted if prematurely terminated.
    The HTTP server must support byte ranges.

    :param url: the URL to download
    :param target_dir: the target path for the file
    :param filename: the filename of the downloaded file
    :param md5sum: the MD5 checksum to match

    :return: the filepath of the downloaded file
    """
    with ImportExtensions(required=True):
        import requests

    def _download(url, target, resume_byte_pos: int = None):
        resume_header = (
            {'Range': f'bytes={resume_byte_pos}-'} if resume_byte_pos else None
        )

        try:
            r = requests.get(url, stream=True, headers=resume_header)
        except requests.exceptions.RequestException as e:
            raise e

        block_size = 1024
        mode = 'ab' if resume_byte_pos else 'wb'

        with target.open(mode=mode) as f:
            for chunk in r.iter_content(32 * block_size):
                f.write(chunk)

    if filename is None:
        filename = url.split('/')[-1]
    filepath = target_dir / filename

    _resume_byte_pos = None
    if filepath.exists():
        if md5sum and md5file(filepath) == md5sum:
            return filepath
        head_info = requests.head(url)
        file_size_online = int(head_info.headers.get('content-length', 0))
        file_size_offline = filepath.stat().st_size
        if file_size_online > file_size_offline:
            _resume_byte_pos = file_size_offline
    _download(url, filepath, _resume_byte_pos)

    if md5sum and not md5file(filepath) == md5sum:
        raise RuntimeError('MD5 checksum failed.')

    return filepath



import os
import sys
import time
from platform import uname
from typing import Dict, TYPE_CHECKING

import requests

from jina.helper import colored, random_port
from jina import __default_host__, __docker_host__

from .base import BaseStore
from ..models import DaemonID
from ..helper import id_cleaner
from ..dockerize import Dockerizer
from ..excepts import Runtime400Exception
from ..models.containers import (
    ContainerArguments,
    ContainerItem,
    ContainerMetadata,
    ContainerStoreStatus,
)

if TYPE_CHECKING:
    from pydantic import BaseModel


class ContainerStore(BaseStore):
    """A Store of Containers spawned by daemon"""

    _kind = 'container'
    _status_model = ContainerStoreStatus

    def _add(self, *args, **kwargs):
        """Implements jina object creation in `mini-jinad`

        .. #noqa: DAR101"""
        raise NotImplementedError

    def _update(self, *args, **kwargs):
        """Implements jina object update in `mini-jinad`

        .. #noqa: DAR101"""
        raise NotImplementedError

    def _delete(self, *args, **kwargs):
        """Implements jina object termination in `mini-jinad`

        .. #noqa: DAR101"""
        raise NotImplementedError

    @property
    def ready(self) -> bool:
        """Check if the container with mini-jinad is alive

        :return: True if mini-jinad is ready"""
        for _ in range(20):
            try:
                r = requests.get(f'{self.host}/')
                if r.status_code == requests.codes.ok:
                    self._logger.success(
                        f'connected to {self.host} to create a {self._kind}'
                    )
                    return True
            except Exception:
                time.sleep(0.5)
                continue
        self._logger.error(f'couldn\'t reach container at {self.host} after 10secs')
        return False

    @BaseStore.dump
    def add(
        self,
        id: DaemonID,
        workspace_id: DaemonID,
        params: 'BaseModel',
        ports: Dict,
        **kwargs,
    ):
        """Add a container to the store

        :param id: id of the container
        :param workspace_id: workspace id where the container lives
        :param params: pydantic model representing the args for the container
        :param ports: ports to be mapped to local
        :param kwargs: keyword args
        :raises KeyError: if workspace_id doesn't exist in the store
        :raises Runtime400Exception: if container creation fails
        :return: id of the container
        """
        try:
            from . import workspace_store

            if workspace_id not in workspace_store:
                raise KeyError(f'{workspace_id} not found in workspace store')

            self.minid_port = random_port()
            # NOTE: jinad when running inside a container needs to access other containers via dockerhost
            # mac/wsl: this would work as is, as dockerhost is accessible.
            # linux: this would only work if we start jinad passing extra_hosts.
            # check if we actually are in docker, needed for unit tests
            # if not docker, use localhost
            if (
                sys.platform == 'linux'
                and 'microsoft' not in uname().release
                and not os.path.exists('/.dockerenv')
            ):
                self.host = f'http://localhost:{self.minid_port}'
            else:
                self.host = f'http://{__docker_host__}:{self.minid_port}'

            self.params = params.dict(exclude={'log_config'})
            # NOTE: `command` is appended to already existing entrypoint, hence removed the prefix `jinad`
            # NOTE: Important to set `workspace_id` here as this gets set in jina objects in the container
            command = f'--port-expose {self.minid_port} --mode {self._kind} --workspace-id {workspace_id.jid}'
            ports.update({f'{self.minid_port}/tcp': self.minid_port})

            container, network, ports = Dockerizer.run(
                workspace_id=workspace_id, container_id=id, command=command, ports=ports
            )
            if not self.ready:
                raise Runtime400Exception(
                    f'{id.type.title()} creation failed, couldn\'t reach the container at {self.host} after 10secs'
                )
            object = self._add(**kwargs)
        except Exception as e:
            self._logger.error(f'got an error while creating the {self._kind}: \n{e}')
            if id in Dockerizer.containers:
                self._logger.info(f'removing container {id_cleaner(container.id)}')
                Dockerizer.rm_container(container.id)
            raise
        else:
            self[id] = ContainerItem(
                metadata=ContainerMetadata(
                    container_id=id_cleaner(container.id),
                    container_name=container.name,
                    image_id=id_cleaner(container.image.id),
                    network=network,
                    ports=ports,
                    host=self.host,
                ),
                arguments=ContainerArguments(
                    command=command,
                    object=object,
                ),
                workspace_id=workspace_id,
            )
            self._logger.success(
                f'{colored(id, "green")} is added to workspace {colored(workspace_id, "green")}'
            )

            del self.host
            workspace_store[workspace_id].metadata.managed_objects.add(id)
            return id

    @BaseStore.dump
    def delete(self, id: DaemonID, **kwargs) -> None:
        """Delete a container from the store

        :param id: id of the container
        :param kwargs: keyword args
        :raises KeyError: if id doesn't exist in the store
        """
        if id not in self:
            raise KeyError(f'{colored(id, "red")} not found in store.')

        self._delete(host=self[id].metadata.host)
        workspace_id = self[id].workspace_id
        del self[id]
        from . import workspace_store

        workspace_store[workspace_id].metadata.managed_objects.remove(id)
        self._logger.success(f'{colored(id, "green")} is released from the store.')



"""Argparser module for hub push"""
from ..helper import add_arg_group


def mixin_hub_pull_parser(parser):
    """Add the arguments for hub pull to the parser
    :param parser: the parser configure
    """
    gp = add_arg_group(parser, title='Pull')
    gp.add_argument(
        'uri',
        type=str,
        help='The URI of the executor to download (e.g., jinahub(+docker)://dummy_executor)',
    )



from collections import defaultdict
from datetime import datetime
from enum import Enum
from types import SimpleNamespace
from typing import Callable, Dict, Any, Optional, List, Union, Tuple

from google.protobuf.descriptor import Descriptor, FieldDescriptor
from google.protobuf.json_format import MessageToDict
from google.protobuf.pyext.cpp_message import GeneratedProtocolMessageType
from pydantic import Field, BaseModel, BaseConfig, create_model, root_validator

from .....proto import jina_pb2
from .....proto.jina_pb2 import (
    DenseNdArrayProto,
    NdArrayProto,
    SparseNdArrayProto,
    NamedScoreProto,
    DocumentProto,
    RouteProto,
    EnvelopeProto,
    StatusProto,
    MessageProto,
    RequestProto,
)
from .....types.document import Document

PROTO_TO_PYDANTIC_MODELS = SimpleNamespace()
PROTOBUF_TO_PYTHON_TYPE = {
    FieldDescriptor.TYPE_INT32: int,
    FieldDescriptor.TYPE_INT64: int,
    FieldDescriptor.TYPE_UINT32: int,
    FieldDescriptor.TYPE_UINT64: int,
    FieldDescriptor.TYPE_SINT32: int,
    FieldDescriptor.TYPE_SINT64: int,
    FieldDescriptor.TYPE_BOOL: bool,
    FieldDescriptor.TYPE_FLOAT: float,
    FieldDescriptor.TYPE_DOUBLE: float,
    FieldDescriptor.TYPE_FIXED32: float,
    FieldDescriptor.TYPE_FIXED64: float,
    FieldDescriptor.TYPE_SFIXED32: float,
    FieldDescriptor.TYPE_SFIXED64: float,
    FieldDescriptor.TYPE_BYTES: bytes,
    FieldDescriptor.TYPE_STRING: str,
    FieldDescriptor.TYPE_ENUM: Enum,
    FieldDescriptor.TYPE_MESSAGE: None,
}


class CustomConfig(BaseConfig):
    """Pydantic config for Camel case and enum handling"""

    use_enum_values = True
    allow_population_by_field_name = True


def _get_oneof_validator(oneof_fields: List, oneof_key: str) -> Callable:
    """
    Pydantic root validator (pre) classmethod generator to confirm only one oneof field is passed

    :param oneof_fields: list of field names for oneof
    :type oneof_fields: List
    :param oneof_key: oneof key
    :type oneof_key: str
    :return: classmethod for validating oneof fields
    """

    def oneof_validator(cls, values):
        if len(set(oneof_fields).intersection(set(values))) > 1:
            raise ValueError(
                f'only one field among {oneof_fields} can be set for key {oneof_key}!'
            )
        return values

    oneof_validator.__qualname__ = 'validate_' + oneof_key
    return root_validator(pre=True, allow_reuse=True)(oneof_validator)


def _get_oneof_setter(oneof_fields: List, oneof_key: str) -> Callable:
    """
    Pydantic root validator (post) classmethod generator to set the oneof key

    :param oneof_fields: list of field names for oneof
    :type oneof_fields: List
    :param oneof_key: oneof key
    :type oneof_key: str
    :return: classmethod for setting oneof fields in Pydantic models
    """

    def oneof_setter(cls, values):
        for oneof_field in oneof_fields:
            if (
                oneof_field in values
                and values[oneof_field] == cls.__fields__[oneof_field].default
            ):
                values.pop(oneof_field)
        return values

    oneof_setter.__qualname__ = 'set_' + oneof_key
    return root_validator(pre=False, allow_reuse=True)(oneof_setter)


def _get_tags_updater() -> Callable:
    """
    Pydantic root validator (pre) classmethod generator to update tags

    :return: classmethod for updating tags in DocumentProto Pydantic model
    """

    def tags_updater(cls, values):
        extra_fields = {k: values[k] for k in set(values).difference(cls.__fields__)}
        if extra_fields:
            if 'tags' not in values:
                values['tags'] = cls.__fields__['tags'].default
            if isinstance(values['tags'], Dict):
                values['tags'].update({i: j for i, j in extra_fields.items()})
        return values

    return root_validator(pre=True, allow_reuse=True)(tags_updater)


def protobuf_to_pydantic_model(
    protobuf_model: Union[Descriptor, GeneratedProtocolMessageType]
) -> BaseModel:
    """
    Converts Protobuf messages to Pydantic model for jsonschema creation/validattion

    ..note:: Model gets assigned in the global Namespace :data:PROTO_TO_PYDANTIC_MODELS

    :param protobuf_model: message from jina.proto file
    :type protobuf_model: Union[Descriptor, GeneratedProtocolMessageType]
    :return: Pydantic model
    """

    all_fields = {}
    camel_case_fields = {}  # {"random_string": {"alias": "randomString"}}
    oneof_fields = defaultdict(list)
    oneof_field_validators = {}

    if isinstance(protobuf_model, Descriptor):
        model_name = protobuf_model.name
        protobuf_fields = protobuf_model.fields
    elif isinstance(protobuf_model, GeneratedProtocolMessageType):
        model_name = protobuf_model.DESCRIPTOR.name
        protobuf_fields = protobuf_model.DESCRIPTOR.fields

    if model_name in vars(PROTO_TO_PYDANTIC_MODELS):
        return PROTO_TO_PYDANTIC_MODELS.__getattribute__(model_name)

    for f in protobuf_fields:
        field_name = f.name
        camel_case_fields[field_name] = {'alias': f.camelcase_name}

        field_type = PROTOBUF_TO_PYTHON_TYPE[f.type]
        default_value = f.default_value

        if f.containing_oneof:
            # Proto Field type: oneof
            # NOTE: oneof fields are handled as a post-processing step
            oneof_fields[f.containing_oneof.name].append(field_name)

        if field_type is Enum:
            # Proto Field Type: enum
            enum_dict = {}
            for enum_field in f.enum_type.values:
                enum_dict[enum_field.name] = enum_field.number
            field_type = Enum(f.enum_type.name, enum_dict)

        if f.message_type:

            if f.message_type.name == 'Struct':
                # Proto Field Type: google.protobuf.Struct
                field_type = Dict
                default_value = {}

            elif f.message_type.name == 'Timestamp':
                # Proto Field Type: google.protobuf.Timestamp
                field_type = datetime
                default_value = datetime.now()

            else:
                # Proto field type: Proto message defined in jina.proto
                if f.message_type.name == model_name:
                    # Self-referencing models
                    field_type = model_name
                else:
                    # This field_type itself a Pydantic model
                    field_type = protobuf_to_pydantic_model(f.message_type)
                    PROTO_TO_PYDANTIC_MODELS.__setattr__(model_name, field_type)

        if f.label == FieldDescriptor.LABEL_REPEATED:
            field_type = List[field_type]

        all_fields[field_name] = (field_type, Field(default=default_value))

    # Post-processing (Handle oneof fields)
    for oneof_k, oneof_v_list in oneof_fields.items():
        oneof_field_validators[f'oneof_validator_{oneof_k}'] = _get_oneof_validator(
            oneof_fields=oneof_v_list, oneof_key=oneof_k
        )
        oneof_field_validators[f'oneof_setter_{oneof_k}'] = _get_oneof_setter(
            oneof_fields=oneof_v_list, oneof_key=oneof_k
        )

    if model_name == 'DocumentProto':
        oneof_field_validators['tags_validator'] = _get_tags_updater()

    CustomConfig.fields = camel_case_fields
    model = create_model(
        model_name,
        **all_fields,
        __config__=CustomConfig,
        __validators__=oneof_field_validators,
    )
    model.update_forward_refs()
    PROTO_TO_PYDANTIC_MODELS.__setattr__(model_name, model)
    return model


for proto in (
    DenseNdArrayProto,
    NdArrayProto,
    SparseNdArrayProto,
    NamedScoreProto,
    DocumentProto,
    RouteProto,
    EnvelopeProto,
    StatusProto,
    MessageProto,
    RequestProto,
):
    protobuf_to_pydantic_model(proto)


def _to_camel_case(snake_str: str) -> str:
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])


class JinaStatusModel(BaseModel):
    """Pydantic BaseModel for Jina status, used as the response model in REST app."""

    jina: Dict
    envs: Dict
    used_memory: str

    class Config:
        alias_generator = _to_camel_case
        allow_population_by_field_name = True


def _get_example_data():
    _example = jina_pb2.DocumentArrayProto()
    d0 = Document(id='🐲', tags={'guardian': 'Azure Dragon', 'position': 'East'})
    d1 = Document(id='🐦', tags={'guardian': 'Vermilion Bird', 'position': 'South'})
    d2 = Document(id='🐢', tags={'guardian': 'Black Tortoise', 'position': 'North'})
    d3 = Document(id='🐯', tags={'guardian': 'White Tiger', 'position': 'West'})
    d0.chunks.append(d1)
    d0.chunks[0].chunks.append(d2)
    d0.matches.append(d3)
    _example.docs.extend([d0.proto, d1.proto, d2.proto, d3.proto])
    return MessageToDict(
        _example,
        including_default_value_fields=True,
    )['docs']


class JinaRequestModel(BaseModel):
    """
    Jina HTTP request model.
    """

    data: Optional[
        Union[
            List[PROTO_TO_PYDANTIC_MODELS.DocumentProto],
            List[Dict[str, Any]],
            List[str],
            List[bytes],
            List[
                Tuple[
                    PROTO_TO_PYDANTIC_MODELS.DocumentProto,
                    PROTO_TO_PYDANTIC_MODELS.DocumentProto,
                ]
            ],
            List[Tuple[Dict[str, Any], Dict[str, Any]]],
            List[Tuple[str, str]],
            List[Tuple[bytes, bytes]],
        ]
    ] = Field(
        None,
        example=_get_example_data(),
        description='Data to send, a list of dict/string/bytes that can be converted into a list of `Document` objects',
    )
    target_peapod: Optional[str] = Field(
        None,
        examples='pod1/*',
        description='A regex string represent the certain peas/pods request targeted.',
    )
    parameters: Optional[Dict] = Field(
        None,
        example={'top_k': 3, 'model': 'bert'},
        description='A dictionary of parameters to be sent to the executor.',
    )

    class Config:
        alias_generator = _to_camel_case
        allow_population_by_field_name = True


class JinaResponseModel(BaseModel):
    """
    Jina HTTP Response model. Only `request_id` and `data` are preserved.
    """

    class DataRequestModel(BaseModel):
        docs: Optional[List[Dict[str, Any]]] = None
        groundtruths: Optional[List[Dict[str, Any]]] = None

    request_id: str = Field(
        ...,
        example='b5110ed9-1954-4a3d-9180-0795a1e0d7d8',
        description='The id given by Jina service',
    )
    data: Optional[DataRequestModel] = Field(None, description='Returned Documents')

    class Config:
        alias_generator = _to_camel_case
        allow_population_by_field_name = True


class JinaEndpointRequestModel(JinaRequestModel):
    """
    Jina HTTP request model that allows customized endpoint.
    """

    exec_endpoint: str = Field(
        ...,
        example='/foo',
        description='The endpoint string, by convention starts with `/`. '
        'All executors bind with `@requests(on="/foo")` will receive this request.',
    )



"""Module containing the Base Client for Jina."""
import abc
import argparse
import inspect
import os
from abc import ABC
from typing import Callable, Union, Optional, Iterator, AsyncIterator

from ..request import GeneratorSourceType
from ...excepts import BadClientInput, ValidationError
from ...helper import typename, ArgNamespace
from ...logging.logger import JinaLogger
from ...logging.predefined import default_logger
from ...parsers import set_client_cli_parser
from ...types.request import Request, Response

InputType = Union[GeneratorSourceType, Callable[..., GeneratorSourceType]]
CallbackFnType = Optional[Callable[[Response], None]]


class BaseClient(ABC):
    """A base client for connecting to the Flow Gateway.

    :param args: the Namespace from argparse
    :param kwargs: additional parameters that can be accepted by client parser
    """

    def __init__(self, args: Optional['argparse.Namespace'] = None, **kwargs):
        if args and isinstance(args, argparse.Namespace):
            self.args = args
        else:
            self.args = ArgNamespace.kwargs2namespace(kwargs, set_client_cli_parser())
        self.logger = JinaLogger(self.__class__.__name__, **vars(self.args))

        if not self.args.proxy and os.name != 'nt':
            # (Han 2020 12.12): gRPC channel is over HTTP2 and it does not work when we have proxy
            # as many enterprise users are behind proxy, a quick way to
            # surpass it is by temporally unset proxy. Please do NOT panic as it will NOT
            # affect users os-level envs.
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')
        self._inputs = None

    @staticmethod
    def check_input(inputs: Optional[InputType] = None, **kwargs) -> None:
        """Validate the inputs and print the first request if success.

        :param inputs: the inputs
        :param kwargs: keyword arguments
        """

        if inputs is None:
            # empty inputs is considered as valid
            return

        if hasattr(inputs, '__call__'):
            # it is a function
            inputs = inputs()

        kwargs['data'] = inputs
        kwargs['exec_endpoint'] = '/'

        if inspect.isasyncgenfunction(inputs) or inspect.isasyncgen(inputs):
            raise ValidationError(
                'checking the validity of an async generator is not implemented yet'
            )

        try:
            from ..request import request_generator

            r = next(request_generator(**kwargs))
            if isinstance(r, Request):
                default_logger.success(f'inputs is valid')
            else:
                raise TypeError(f'{typename(r)} is not a valid Request')
        except Exception as ex:
            default_logger.error(f'inputs is not valid!')
            raise BadClientInput from ex

    def _get_requests(
        self, **kwargs
    ) -> Union[Iterator['Request'], AsyncIterator['Request']]:
        """
        Get request in generator.

        :param kwargs: Keyword arguments.
        :return: Iterator of request.
        """
        _kwargs = vars(self.args)
        _kwargs['data'] = self.inputs
        # override by the caller-specific kwargs
        _kwargs.update(kwargs)

        if inspect.isasyncgen(self.inputs):
            from ..request.asyncio import request_generator

            return request_generator(**_kwargs)
        else:
            from ..request import request_generator

            return request_generator(**_kwargs)

    @property
    def inputs(self) -> InputType:
        """
        An iterator of bytes, each element represents a Document's raw content.

        ``inputs`` defined in the protobuf

        :return: inputs
        """
        return self._inputs

    @inputs.setter
    def inputs(self, bytes_gen: InputType) -> None:
        """
        Set the input data.

        :param bytes_gen: input type
        """
        if hasattr(bytes_gen, '__call__'):
            self._inputs = bytes_gen()
        else:
            self._inputs = bytes_gen

    @abc.abstractmethod
    async def _get_results(
        self,
        inputs: InputType,
        on_done: Callable,
        on_error: Callable = None,
        on_always: Callable = None,
        **kwargs,
    ):
        ...

    @property
    def client(self) -> 'BaseClient':
        """Return the client object itself

        :return: the Client object
        """
        return self



import time
import asyncio
import argparse
from typing import Optional

from ....enums import RemoteWorkspaceState
from ..asyncio.base import AsyncZMQRuntime
from ....helper import cached_property, colored
from ....excepts import DaemonConnectivityError
from .client import PeaDaemonClient, WorkspaceDaemonClient


class JinadRuntime(AsyncZMQRuntime):
    """Runtime procedure for Jinad."""

    def __init__(self, args: 'argparse.Namespace', ctrl_addr: str):
        super().__init__(args, ctrl_addr)
        # Need the `proper` control address to send `activate` and `deactivate` signals, from the pea in the `main`
        # process.
        self.host = args.host
        self.port_expose = args.port_expose
        # NOTE: args.timeout_ready is always set to -1 for JinadRuntime so that wait_for_success doesn't fail in Pea,
        # so it can't be used for Client timeout.
        self.workspace_api = WorkspaceDaemonClient(
            host=self.host,
            port=self.port_expose,
            logger=self.logger,
        )
        self.pea_api = PeaDaemonClient(
            host=self.host,
            port=self.port_expose,
            logger=self.logger,
        )
        # Uploads Pea context to remote & Creates remote Pea using :class:`PeaDaemonClient`
        if self._remote_id:
            self.logger.success(
                f'created a remote {self.pea_api.kind}: {colored(self._remote_id, "cyan")}'
            )

    async def async_run_forever(self):
        """
        Streams log messages using websocket from remote server
        """
        self._logging_task = asyncio.create_task(
            self._sleep_forever()
            if self.args.quiet_remote_logs
            else self.pea_api.logstream(self._remote_id)
        )

    async def async_cancel(self):
        """
        Cancels the logging task
        """
        self._logging_task.cancel()

    def teardown(self):
        """
        Terminates the remote Workspace/Pod/Pea using :class:`JinadAPI`
        """
        self.pea_api.delete(id=self._remote_id)
        # TODO: don't fail if workspace deletion fails. all peas would make this call. can be optimized
        self.workspace_api.delete(id=self.args.workspace_id)
        super().teardown()

    def create_workspace(self):
        """Create a workspace on remote (includes file upload & docker build)

        :raises DaemonConnectivityError: if remote daemon is not reachable
        :raises RuntimeError: if workspace creation fails
        """
        if not self.workspace_api.alive:
            self.logger.error(
                f'fail to connect to the daemon at {self.host}:{self.port_expose}, please check:\n'
                f'- is there a typo in {self.host}?\n'
                f'- on {self.host}, are you running `docker run --network host jinaai/jina:latest-daemon`?\n'
                f'- on {self.host}, have you set the security policy to public for all required ports?\n'
                f'- on local, are you behind VPN or proxy?'
            )
            raise DaemonConnectivityError

        sleep = 2
        retries = 100
        for retry in range(retries):
            workspace_status = self.workspace_api.get(id=self.args.workspace_id)
            if not workspace_status:
                raise DaemonConnectivityError
            state = workspace_status.get('state', None)
            if not state:
                self.logger.info(
                    f'creating workspace {colored(self.args.workspace_id, "cyan")} on remote. This might take some time.'
                )
                self.workspace_api.post(
                    dependencies=self.args.upload_files,
                    workspace_id=self.args.workspace_id,
                )
            elif state in [
                RemoteWorkspaceState.PENDING,
                RemoteWorkspaceState.CREATING,
                RemoteWorkspaceState.UPDATING,
            ]:
                if retry % 10 == 0:
                    self.logger.info(
                        f'workspace {self.args.workspace_id} is getting created on remote. waiting..'
                    )
                time.sleep(sleep)
            elif state == RemoteWorkspaceState.ACTIVE:
                self.logger.success(
                    f'successfully created a remote workspace: {colored(self.args.workspace_id, "cyan")}'
                )
                break
            else:
                raise RuntimeError(f'remote workspace creation failed')

    @cached_property
    def _remote_id(self) -> Optional[str]:
        """Creates a workspace & a pea on remote

        :return: id of rempte pea
        """
        self.create_workspace()
        pea_id = self.pea_api.post(self.args)
        if not pea_id:
            raise RuntimeError('remote pea creation failed')
        return pea_id

    async def _sleep_forever(self):
        """Sleep forever, no prince will come."""
        await asyncio.sleep(1e10)



"""Helper functions for clients in Jina."""

from functools import wraps
from typing import Callable

from ..excepts import BadClientCallback
from ..helper import colored
from ..importer import ImportExtensions
from ..logging.logger import JinaLogger
from ..proto import jina_pb2
from ..types.request import Response


def pprint_routes(resp: 'Response', stack_limit: int = 3):
    """Pretty print routes with :mod:`prettytable`, fallback to :func:`print`.

    :param resp: the :class:`Response` object
    :param stack_limit: traceback limit
    """
    from textwrap import fill

    routes = resp.routes

    header = [colored(v, attrs=['bold']) for v in ('Pod', 'Time', 'Exception')]

    with ImportExtensions(required=False):
        from prettytable import PrettyTable, ALL

        table = PrettyTable(field_names=header, align='l', hrules=ALL)
        add_row = table.add_row
        visualize = print

    for route in routes:
        status_icon = '🟢'
        if route.status.code == jina_pb2.StatusProto.ERROR:
            status_icon = '🔴'
        elif route.status.code == jina_pb2.StatusProto.ERROR_CHAINED:
            status_icon = '⚪'

        add_row(
            [
                f'{status_icon} {route.pod}',
                f'{route.start_time.ToMilliseconds() - routes[0].start_time.ToMilliseconds()}ms',
                fill(
                    ''.join(route.status.exception.stacks[-stack_limit:]),
                    width=50,
                    break_long_words=False,
                    replace_whitespace=False,
                ),
            ]
        )

    visualize(table)


def _safe_callback(func: Callable, continue_on_error: bool, logger) -> Callable:
    @wraps(func)
    def _arg_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            err_msg = f'uncaught exception in callback {func.__name__}(): {ex!r}'
            if continue_on_error:
                logger.error(err_msg)
            else:
                raise BadClientCallback(err_msg) from ex

    return _arg_wrapper


def callback_exec(
    response,
    on_done: Callable,
    on_error: Callable,
    on_always: Callable,
    continue_on_error: bool,
    logger: JinaLogger,
) -> None:
    """Execute the callback with the response.

    :param response: the response
    :param on_done: the on_done callback
    :param on_error: the on_error callback
    :param on_always: the on_always callback
    :param continue_on_error: whether to continue on error
    :param logger: a logger instance
    """
    if on_error and response.status.code >= jina_pb2.StatusProto.ERROR:
        _safe_callback(on_error, continue_on_error, logger)(response)
    elif on_done:
        _safe_callback(on_done, continue_on_error, logger)(response)
    if on_always:
        _safe_callback(on_always, continue_on_error, logger)(response)



import os
from pathlib import Path

DEFAULT_JINA_HUB_ROOT = Path.home().joinpath('.jina', 'hub-packages')
DEFAULT_JINA_HUB_CACHE_DIR = Path.home().joinpath('.cache', 'jina')

JINA_HUB_ROOT = Path(os.environ.get('JINA_HUB_ROOT', DEFAULT_JINA_HUB_ROOT))
JINA_HUB_ROOT.mkdir(parents=True, exist_ok=True)

JINA_HUB_CACHE_DIR = Path(
    os.environ.get('JINA_HUB_CACHE_DIR', DEFAULT_JINA_HUB_CACHE_DIR)
)
JINA_HUB_CACHE_DIR.mkdir(parents=True, exist_ok=True)
